# -*- coding: utf-8 -*-

from decimal import Decimal
from jinja2 import Environment, PackageLoader
from marcellus.ficherocsv import FicheroCSV
from marcellus.util import default_fmt_float
from marcellus.xlsreport import XLSReport
from sqlalchemy.types import INTEGER, NUMERIC, DATE, TIME, BOOLEAN, BIGINT, \
    DATETIME, TIMESTAMP
import datetime as dt
import re
import sys

COLUMNAS          = 'columnas'
COUNT             = 'count'
LIMITE_RESULTADOS = 'limite_resultados'
DATOS             = 'datos'
TOTALES           = 'totales'


def tocsv(e):
    def _tocsv(func):
        def __tocsv(*args, **kwargs):
            return func(*args, **kwargs).to_csv(encoding=e)

        return __tocsv

    return _tocsv


def tostring(func):
    def __tostring(*args, **kwargs):
        return func(*args, **kwargs).tostring()

    return __tostring


def tojson(func):
    def __tojson(*args, **kwargs):
        return func(*args, **kwargs).to_json()

    return __tojson


def remove_specials(texto):
    return re.sub(r'[^a-zA-Z0-9_]', '_', texto)


class DataSetRowIterator(object):

    def __init__(self, row):
        self.row = row
        self.i = 0
        self.l = len(self.row.cols)

    def __iter__(self):
        return self

    def next(self):
        if self.i < self.l:
            item = self.row[self.i]
            self.i += 1

            return item
        else:
            raise StopIteration

    def __next__(self):
        return self.next()


class DataSetRow(object):

    cols = []

    def __init__(self, **kw):
        self.attr = kw

    def __str__(self):
        result = []
        for k, v in self.attr.iteritems():
            result.append('{0}={1}'.format(k, v))

        return ', '.join(result)

    def __getattr__(self, name):
        if name in self.attr:
            return self.attr[name]
        else:
            raise Exception('<DataSet>: member "%s" does not exist' % name)

    def __iter__(self):
        return DataSetRowIterator(self)

    def __getitem__(self, key):
        if isinstance(key, int):
            k = self.cols[key]
            if sys.version_info < (2, 6, 6):
                if isinstance(k, unicode):
                    k = k.encode('utf-8')

            if k in self.attr:
                return self.attr[k]

            else:
                return None

        elif key in self.attr:
            return self.attr[key]

    def __setitem__(self, key, value):
        if isinstance(key, int):
            k = self.cols[key]
            if sys.version_info < (2, 6, 6):
                if isinstance(k, unicode):
                    k = k.encode('utf-8')

            self.attr[k] = value

        else:
            if sys.version_info < (2, 6, 6):
                if isinstance(key, unicode):
                    key = key.encode('utf-8')

            self.attr[key] = value


class DataSet(object):

    date_fmt = '%d/%m/%Y'
    time_fmt = '%H:%M'
    datetime_fmt = '%d/%m/%Y %H:%M:%S'
    true_const = True
    false_const = False
    int_fmt = '%d'

    def float_fmt(self, value):
        return default_fmt_float(value)

    def __init__(self, cols, types=None, limite=None, totales=None, **fmt):

        self.cols = []
        self.labels = []
        self.types = []

        # date
        if 'date_fmt' in fmt:
            self.date_fmt = fmt['date_fmt']

        # time
        if 'time_fmt' in fmt:
            self.time_fmt = fmt['time_fmt']

        # datetime
        if 'datetime_fmt' in fmt:
            self.datetime_fmt = fmt['datetime_fmt']

        # int
        if 'int_fmt' in fmt:
            self.int_fmt = fmt['int_fmt']

        # float
        if 'float_fmt' in fmt:
            self.float_fmt = fmt['float_fmt']

        for col in cols:
            self.append_col(col)

        self.limite_resultados = limite
        self.data = []
        self.count = 0
        self.totales = totales

    def append_col(self, col):
        if isinstance(col, tuple):
            # (<col>, <etiqueta>, <tipo>,)
            self.cols.append(col[0])
            self.labels.append(col[1] or col[0])
            self.types.append(col[2])

        else:
            self.cols.append(col)
            self.labels.append(col)
            self.types.append('')

    def init_totales(self):
        if self.totales:
            totales = dict(zip(self.totales, [None] * len(self.totales)))

        else:
            totales = {}

        return totales

    def acumular(self, acumulado, valor):
        return (acumulado or 0) + (valor or 0)

    def append(self, dato=None):
        """
        IN
          dato
            <list>
            <dict>
            <DataSetRow>

        OUT
          ???
        """

        if isinstance(dato, DataSetRow):
            self.data.append(dato)

        elif isinstance(dato, dict):
            # compare keys and columns
            d = {}
            for k, v in dato.iteritems():
                if k not in self.cols:
                    raise Exception('The key "{0}" is not correct'.format(k))

                if sys.version < (2, 6):
                    if isinstance(k, unicode):
                        k = k.encode('utf-8')

                d[k] = v

            self.append(DataSetRow(**d))

        elif isinstance(dato, DataSet):
            for d in dato:
                self.append(d)

        elif isinstance(dato, list) or hasattr(dato, '__iter__'):
            d = {}
            for col, v in zip(self.cols, dato):
                if sys.version_info < (2, 6, 6):
                    if isinstance(col, unicode):
                        col = col.encode('utf-8')

                d[col] = v

            self.append(DataSetRow(**d))

        else:
            self.append([None] * len(self.cols))

        self.data[-1].cols = self.cols
        return self.data[-1]

    def __iter__(self):
        return iter(self.data)

    def reversed(self):
        """Devuelve un iterador para recorrer la lista del final al principio

        Para un DataSet de N valores

        item0
        item1
        ...
        itemN-1

        devolver

        [
         itemN-1,
         ...
         item1,
         item0
        ]
        """
        return reversed(self.data)

    def ereversed(self):
        """
        Devuelve un iterador enumerado para recorrer la lista del final al
        principio.

        Para un DataSet de N valores

        item0
        item1
        ...
        itemN-1

        devolver

        [
         (N-1, itemN-1)
         ...
         (1, item1)
         (0, item0)
        ]

        """
        return zip(range(len(self) - 1, -1, -1), self.reversed())

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index]

    def __str__(self):
        return self.to_str()

    def getlabel(self, i):
        if not isinstance(i, int):
            i = self.cols.index(i)

        return self.labels[i]

    def gettype(self, i):
        if not isinstance(i, int):
            i = self.cols.index(i)

        return self.types[i]

    def _format(self, fmt, value):

        if isinstance(fmt, (str, unicode,)):
            s = unicode(fmt % value)

        elif hasattr(fmt, '__call__'):
            s = unicode(fmt(value))

        else:
            s = unicode(value or '')

        return s

    def _formatdt(self, fmt, value):

        if isinstance(fmt, (str, unicode,)):

            if dt.date == type(value) and value >= dt.date(1900, 1, 1) or dt.datetime == type(value) and \
                    value >= dt.datetime(1900, 1, 1) or isinstance(value, dt.time):
                s = unicode(value.strftime(fmt))

            else:
                def _strftime(d, fmt_):
                    if isinstance(d, (dt.datetime, dt.time)):
                        fmt_ = fmt_.replace('%H', '%.2d' % d.hour).\
                            replace('%M', '%.2d' % d.minute).\
                            replace('%S', '%.2d' % d.second)

                    return fmt_.replace('%d', '%.2d' % d.day). \
                        replace('%m', '%.2d' % d.month). \
                        replace('%Y', '%.4d' % d.year)

                s = unicode(_strftime(value, fmt))

        elif hasattr(fmt, '__call__'):
            s = unicode(fmt(value))

        else:
            s = unicode(value or '')

        return s

    def to_str(self, width=None, fit_width=True):
        """
        Devuelve los datos del DataSet en un cadena con formato tabular.

        |------|------|------|------|
        | (c1) | (c2) | (c3) | (c4) |
        |---------------------------|
        |      |      |      |      |
        |      |      |      |      |
        |      |      |      |      |
        |---------------------------|
        |      |      |      |      | <- totales
        |---------------------------|

        IN
          width <int> (opcional)
          Indica el ancho de las columnas

          fit_width <bool> (por defecto=True)
          Hace las columnas tan anchas como el valor m�s ancho, que no sobrepase
          el ancho fijado por 'width'.
        """

        if fit_width:
            widths = [0] * len(self.cols)

            for i, c in enumerate(self.cols):
                for d in self.data:
                    if len(str(d[c])) > widths[i]:
                        widths[i] = len(unicode(d[c]))

                if len(self.labels[i]) > widths[i]:
                    widths[i] = len(self.labels[i])

                if width and widths[i] > width:
                    widths[i] = width
        else:
            widths = [width or 10] * len(self.cols)

        resultado = ''

        cabecera = '|'
        for i, c in enumerate(self.labels):
            cabecera += unicode(c)[:widths[i]].center(widths[i]) + '|'

        l = len(cabecera) - 2

        resultado += '|' + '-' * l + '|\n'
        resultado += cabecera + '\n'
        resultado += '|' + '-' * l + '|\n'

        totales = self.init_totales()

        for dato in self.data:
            linea = '|'
            for i, c in enumerate(self.cols):
                valor = dato[c]
                if valor is None:
                    valor = u''

                w = widths[i]

                # totalizaciones
                if self.totales:
                    if c in totales:
                        totales[c] = self.acumular(totales[c], valor)

                # float/Decimal
                if isinstance(dato[c], (float, Decimal,)):
                    linea += self._format(self.float_fmt, dato[c])[:w].rjust(w)

                # bool
                elif isinstance(dato[c], bool):
                    linea += unicode((self.true_const if dato[c] else self.false_const)).rjust(w)

                # int/long
                elif isinstance(dato[c], (int, long,)):
                    linea += self._format(self.int_fmt, dato[c])[:w].rjust(w)

                # datetime
                elif isinstance(dato[c], dt.datetime):
                    #linea += unicode(dato[c].strftime('%s %s' % (self.date_fmt, self.time_fmt)))[:w].rjust(w)
                    linea += unicode(self._formatdt(self.datetime_fmt, dato[c]))[:w].rjust(w)

                # date
                elif isinstance(dato[c], dt.date):
                    linea += unicode(self._formatdt(self.date_fmt, dato[c]))[:w].rjust(w)

                # time
                elif isinstance(dato[c], dt.time):
                    #linea += unicode(dato[c].strftime(self.time_fmt))[:w].rjust(w)
                    linea += unicode(self._formatdt(self.time_fmt, dato[c]))[:w].rjust(w)

                else:
                    linea += unicode(valor).replace('\n', '')[:w].ljust(w)

                linea += '|'

            resultado += linea + '\n'

        resultado += '|' + '-' * l + '|\n'

        # línea de totales
        if self.totales:
            linea = '|'
            for i, c in enumerate(self.cols):
                w = widths[i]
                if c in self.totales:
                    linea += self.float_fmt(totales[c])[:w].rjust(w)
                else:
                    linea += ''.rjust(w)

                linea += '|'

            resultado += linea + '\n'
            resultado += '|' + '-' * l + '|\n'

        return resultado

    def to_data(self):
        """
        """

        data = []

        for row in self.data:
            dato = []
            for item in row:

                # datetime
                if isinstance(item, dt.datetime):
                    dato.append(self._formatdt(self.datetime_fmt, item))

                # date
                elif isinstance(item, dt.date):
                    dato.append(self._formatdt(self.date_fmt, item))

                # time
                elif isinstance(item, dt.time):
                    dato.append(self._formatdt(self.time_fmt, item))

                # float, Decimal
                elif isinstance(item, (float, Decimal,)):
                    dato.append(self._format(self.float_fmt, item))

                # bool
                elif isinstance(item, bool):
                    dato.append(self.true_const if item else self.false_const)

                # int, long
                elif isinstance(item, (int, long,)):
                    dato.append(self._format(self.int_fmt, item))

                # unicode
                elif isinstance(item, unicode):
                    dato.append(item)

                # ???
                else:
                    dato.append(str(item or '').decode('utf-8'))

            data.append(dato)

        return data

    def to_csv(self, encoding=None, mostrar_ids=False):
        """
        Devuelve en una cadena con formato CSV el contenido del DataSet.

        IN
          encoding <str> (opcional)

        OUT
          <str> (csv)
        """

        fichero_csv = FicheroCSV()
        fichero_csv.date_fmt = self.date_fmt
        fichero_csv.time_fmt = self.time_fmt
        fichero_csv.true_const = self.true_const
        fichero_csv.false_const = self.false_const
        fichero_csv.float_fmt = self.float_fmt

        columnas = []
        labels = []
        for item, c in enumerate(self.cols):

            label = self.labels[item]
#            if isinstance(label, unicode):
#                label = label.encode('utf-8')

            if not c.startswith('id') or mostrar_ids:
                columnas.append(remove_specials(self.cols[item]))
                labels.append(label)

        fichero_csv.add(labels)

        totales = self.init_totales()

        for dato in self.data:

            # totalizar
            if self.totales:
                for idx, c in enumerate(columnas):
                    if c in totales:
                        totales[c] = self.acumular(totales[c], dato[idx])

            fila = []
            for col in columnas:
                fila.append(dato[col])

            fichero_csv.add(fila)

        # añadir línea con el total
        if self.totales:
            fila_totales = []
            for c in columnas:
                if c in totales:
                    fila_totales.append(self.float_fmt(totales[c]))

                else:
                    fila_totales.append(None)

            fichero_csv.add(fila_totales)

        return fichero_csv.read(encoding)

    def to_xls(self, title, filename, fmt=None):
        pl = PackageLoader('marcellus', 'templates')
        env = Environment(loader=pl)
        template = env.get_template('dstoxls.xml')

        xr = XLSReport(template)

        date_fmt = self.date_fmt.replace('%d', 'dd').\
            replace('%m', 'mm').\
            replace('%Y', 'yyyy')

        time_fmt = self.time_fmt.\
            replace('%H', 'HH').\
            replace('%M', 'MM').\
            replace('%S', 'SS')

        _fmt = dict(date=date_fmt, time=time_fmt)
        _fmt.update(fmt or {})

        params = dict(title=title, ds=self, fmt=_fmt)
        return xr.create(params, filename=filename)

    def order(self, claves):
        """
        IN
          claves <list> [(<col>, <signo>), ...]
            * <col>    Nombre de la columna
            * <signo>  +1 para ordenación ascendente
                       -1 para ordenación descendente

        OUT
          ???
        """

        def ordenar(ordenar_por):
            def _ordenar(function):
                def __ordenar(*args):
                    for o, signo in ordenar_por:
                        new_args = (args[0][o], args[1][o])
                        r = function(*new_args)

                        if r != 0:
                            break
                    return r * signo

                return __ordenar

            return _ordenar

        @ordenar(claves)
        def compara(x, y):
            if x < y:
                return -1

            elif x == y:
                return 0

            else:
                return 1

        self.data.sort(cmp=compara)

    @staticmethod
    def procesar_resultado(session, query, limit=None, pos=None, no_count=False, show_ids=False):
        """
        IN
          session  <sqlalchemy.orm.session.Session>
          query    <>
          limit    <int>
          pos      <int>
          no_count  <bool> = False
          show_ids  <bool> = False

        OUT
          <DataSet>
        """

        cols = []
        col_names = []
        for c in query.columns:
            if show_ids or not (c.name.startswith('id_') or c.name.startswith('_')):
                type_ = ''
                if isinstance(c.type, INTEGER) or isinstance(c.type, BIGINT):
                    type_ = 'int'

                elif isinstance(c.type, NUMERIC):
                    type_ = 'float'

                elif isinstance(c.type, DATE):
                    type_ = 'date'

                elif isinstance(c.type, TIME):
                    type_ = 'time'

                elif isinstance(c.type, DATETIME) or isinstance(c.type, TIMESTAMP):
                    type_ = 'datetime'

                elif isinstance(c.type, BOOLEAN):
                    type_ = 'bool'

                # avoid col name repeat
                name = remove_specials(c.name).lower()
                i = 1
                while name in col_names:
                    name = u'{0}_{1}'.format(name, i)
                    i += 1

                col_names.append(name)

                cols.append((name, c.name, type_,))

        ds = DataSet(cols)
        if not no_count:
            ds.count = session.execute(query).rowcount

        else:
            ds.count = 0

        if limit == 0:
            limit = None

        for fila in session.execute(query.limit(limit).offset(pos)):
            row = []
            for c in query.columns:
                data = fila[c]
                if show_ids or not (c.name.startswith('id_') or c.name.startswith('_')):
                    if isinstance(data, str):
                        data = data.decode('utf-8')

                    if isinstance(data, unicode):
                        data = data.replace(u'\u000B', ' ')

                    row.append(data)

            ds.append(row)

            if no_count:
                ds.count += 1

        return ds

    @staticmethod
    def query(dbs, query, cols, limit=None, pos=None):
        """
        IN
          session  <sqlalchemy.orm.session.Session>
          query    <sqlalchemy.orm.query.Query>
          limit    <int>
          pos      <int>

        OUT
          <DataSet>
        """

        ds = DataSet(cols, limite=limit)
        ds.count = dbs.execute(query).rowcount

        if limit == 0:
            limit = None

        for row in dbs.execute(query.limit(limit).offset(pos)):
            ds.append(row.values())

        return ds



# -*- coding: utf-8 -*-

from marcellus.dataset import DataSet
from sqlalchemy import Table, MetaData, select
from sqlalchemy.sql.expression import and_, Select, or_, alias
from sqlalchemy.types import DATE, TIME, DATETIME, TIMESTAMP, INTEGER, NUMERIC, BOOLEAN, BIGINT
import re
import datetime as dt

_CONSTANT_DATE_TODAY       = 'today'
_CONSTANT_DATE_START_WEEK  = 'start_week'
_CONSTANT_DATE_END_WEEK    = 'end_week'
_CONSTANT_DATE_START_MONTH = 'start_month'
_CONSTANT_DATE_END_MONTH   = 'end_month'
_CONSTANT_DATE_START_YEAR  = 'start_year'
_CONSTANT_DATE_END_YEAR    = 'end_year'
_CONSTANT_DAY              = 'd'
_CONSTANT_WEEK             = 'w'
_CONSTANT_MONTH            = 'm'
_CONSTANT_YEAR             = 'y'


class Busqueda(object):

    def __init__(self, tabla, columnas_trans=None, strtodatef=None):
        self.tabla = tabla
        self.cols = []
        self.types = []
        for col in self.tabla.columns:
            self.cols.append(col.name)
            self.types.append(col.type)

        self.cache_campo = {}

        # string to date conversion function
        self.strtodatef = strtodatef

        if not columnas_trans:
            self.cols_trans = [re.sub(r'[^a-z0-9]*', '', self.remove_accents(col))
                               for col in self.cols]
        else:
            self.cols_trans = columnas_trans

    def no_accents(self, texto):

        resultado = texto
        resultado = re.sub(r'(a|á|à|ä)', '(a|á|à|ä)', resultado)
        resultado = re.sub(r'(e|é|è|ë)', '(e|é|è|ë)', resultado)
        resultado = re.sub(r'(i|í|ì|ï)', '(i|í|ì|ï)', resultado)
        resultado = re.sub(r'(o|ó|ò|ö)', '(o|ó|ò|ö)', resultado)
        resultado = re.sub(r'(u|ú|ù|ü)', '(u|ú|ù|ü)', resultado)

        return resultado

    def remove_accents(self, texto):
        texto = texto.encode('utf-8')

        return texto.lower(). \
            replace('á', 'a'). \
            replace('é', 'e'). \
            replace('í', 'i'). \
            replace('ó', 'o'). \
            replace('ú', 'u'). \
            replace('ü', 'u')

    def get_col(self, campo):

        if campo.lower() in self.cache_campo:
            return self.cache_campo[campo.lower()]

        else:
            col = None
            for ct, c in zip(self.cols_trans, self.cols):

                if ct.startswith(campo.lower()):
                    if not col or col['length'] > len(ct):
                        col = dict(name=c, length=len(ct))

            if col:
                self.cache_campo[campo.lower()] = col['name']
                return col['name']

            else:
                return None

    def process_date_constants(self, termino):

        def _sub(m):
            cons = m.group(1)
            s = m.group(3)
            q = int(m.group(4) or 0)
            t = m.group(5) or _CONSTANT_DAY

            r = None
            if cons == _CONSTANT_DATE_TODAY:
                r = dt.date.today()

            elif cons == _CONSTANT_DATE_START_WEEK:
                wd = dt.date.today().weekday()
                r = dt.date.today() - dt.timedelta(days=wd)

            elif cons == _CONSTANT_DATE_END_WEEK:
                wd = 6 - dt.date.today().weekday()
                r = dt.date.today() + dt.timedelta(days=wd)

            elif cons == _CONSTANT_DATE_START_MONTH:
                today = dt.date.today()
                r = dt.date(today.year, today.month, 1)

            elif cons == _CONSTANT_DATE_END_MONTH:
                today = dt.date.today()
                next_month = today + dt.timedelta(days=30)
                r = dt.date(next_month.year, next_month.month, 1) - dt.timedelta(days=1)

            elif cons == _CONSTANT_DATE_START_YEAR:
                today = dt.date.today()
                r = dt.date(today.year, 1, 1)

            elif cons == _CONSTANT_DATE_END_YEAR:
                today = dt.date.today()
                r = dt.date(today.year, 12, 31)

            if r:
                if s and q:
                    if t == _CONSTANT_DAY:
                        if s == '-':
                            n = -q
                        else:
                            n = q

                        r = r + dt.timedelta(days=n)

                    elif t == _CONSTANT_WEEK:
                        if s == '-':
                            n = -q * 7
                        else:
                            n = q * 7

                        r = r + dt.timedelta(days=n)

                    elif t == _CONSTANT_MONTH:

                        r_ = r

                        if s == '-':
                            n = -1
                        else:
                            n = +1

                        def _next_month(d, inc):
                            year = d.year
                            month = 1
                            if inc > 0:
                                if d.month < 12:
                                    month = d.month + inc

                                elif d.month == 12:
                                    month = 1
                                    year += 1
                            else:
                                if d.month > 1:
                                    month = d.month + inc

                                else:
                                    month = 12
                                    year -= 1

                            return dt.date(year, month, 1)

                        i = 0
                        while abs(i) != q:
                            next_month = _next_month(r_, n)

                            if cons == _CONSTANT_DATE_END_MONTH:
                                day = (_next_month(next_month, 1) - dt.timedelta(days=1)).day

                            else:
                                day = r.day

                            r_ = dt.date(next_month.year, next_month.month, day)

                            i += n

                        r = r_

                    elif t == _CONSTANT_YEAR:
                        if s == '-':
                            n = -q

                        else:
                            n = +q

                        r = dt.date(r.year + n, r.month, r.day)

                return r.strftime('%Y-%m-%d')

            else:
                return m.group(0)

        return re.sub(r'\{\s*(\w+)(\s+([\+\-])\s*(\d+)\s*(\w?))?\s*}', _sub, termino)

    def _get_date(self, s):
        if self.strtodatef:
            try:
                return self.strtodatef(s).strftime('%Y-%m-%d')
            except:
                s_ = self.process_date_constants(s)
                if s != s_:
                    return s_

                return None
        else:
            try:
                return dt.datetime.strptime(s, '%d/%m/%Y').strftime('%Y-%m-%d')
            except ValueError:
                s_ = self.process_date_constants(s)
                if s != s_:
                    return s_

                return None

    def _get_time(self, s):
        m_time = re.search(r'^\d{1,2}:\d{1,2}(:\d{1,2})?$', s)
        if m_time:
            return s

    def _get_datetime(self, s):
        m_datetime = re.search(r'^\d{1,2}/\d{1,2}/\d{4}\s+\d{1,2}:\d{1,2}(:\d{1,2})?$', s)
        if m_datetime:
            fmt = '%d/%m/%Y %H:%M'
            if m_datetime.group(1):
                fmt += ':%S'

            try:
                return dt.datetime.strptime(s, fmt).strftime('%Y-%m-%d %H:%M:%S')
            except ValueError:
                return None

    def _get_number(self, s):
        m_number = re.search(r'^\d+(\.\d+)?$', s)
        if m_number:
            return s

    def process(self, s):

        filters = []
        order_by = []
        terminos = re.findall(r'(!|#|\+|\-)?(\w[\w\s]+\w?|\w)\s*(<>|<=|>=|==|!=|=|<|>)?\s*("[^"]+"|\d+\.\d+|\d+|\w[\w\s]+\w?)?,?', s)
        for termino in terminos:
            operator0 = termino[0]
            field     = termino[1].strip()
            operator  = termino[2].strip()

            # remove " at the begining and the end
            operand = termino[3]
            m_operand = re.search(r'^"([^"]+)"$', operand)
            if m_operand:
                operand = m_operand.group(1)

            c = None
            if operator == '' and operator0 == '':
                for operand in field.split():
                    # buscar en cada columna por el término correspondiente
                    filters_ = []
                    for col in self.tabla.columns:
                        col_name = col.name.encode('utf-8')

                        if col_name == 'id' or col_name.startswith('id_'):
                            continue

                        filter_ = []
                        if isinstance(col.type, DATE) or isinstance(col.type, TIME) or \
                                isinstance(col.type, DATETIME) or isinstance(col.type, TIMESTAMP):
                            filter_.append('"{0}" IS NOT NULL'.format(col_name))

                        elif isinstance(col.type, INTEGER) or isinstance(col.type, NUMERIC) or \
                                isinstance(col.type, BIGINT):
                            filter_.append('"{0}" IS NOT NULL'.format(col_name))

                        else:
                            filter_.append('COALESCE(CAST("{0}" as Text), \'\') != \'\''.format(col_name))

                        args_ = (col_name, self.no_accents(operand))
                        filter_.append("UPPER(CAST(\"{0}\" as Text)) SIMILAR TO UPPER('%{1}%')".format(*args_))

                        filters_.append('({0})'.format(' AND '.join(filter_)))

                    filters.append('({0})'.format('\nOR '.join(filters_)))

            else:
                c = self.get_col(field)

            if not c:
                continue

            col = self.tabla.columns[c]

            if operator0 == '#':
                # DATE, TIME, DATETIME/TIMESTAMP
                if isinstance(col.type, DATE) or isinstance(col.type, TIME) or \
                        isinstance(col.type, DATETIME) or isinstance(col.type, TIMESTAMP):
                    filters.append('"{0}" IS NOT NULL'.format(c.encode('utf-8')))

                # INTEGER, NUMERIC, BIGINT
                elif isinstance(col.type, INTEGER) or isinstance(col.type, NUMERIC) or isinstance(col.type, BIGINT):
                    filters.append('"{0}" IS NOT NULL'.format(c.encode('utf-8')))

                # BOOLEAN
                elif isinstance(col.type, BOOLEAN):
                    filters.append('"{0}" = TRUE'.format(c.encode('utf-8')))

                else:
                    filters.append('TRIM(COALESCE(CAST("{0}" as Text), \'\')) != \'\''.format(c.encode('utf-8')))

            elif operator0 == '!':
                # DATE, TIME, DATETIME/TIMESTAMP
                if isinstance(col.type, DATE) or isinstance(col.type, TIME) or \
                        isinstance(col.type, DATETIME) or isinstance(col.type, TIMESTAMP):
                    filters.append('"{0}" IS NULL'.format(c.encode('utf-8')))

                # INTEGER, NUMERIC, BIGINT
                elif isinstance(col.type, INTEGER) or isinstance(col.type, NUMERIC) or isinstance(col.type, BIGINT):
                    filters.append('"{0}" IS NULL'.format(c.encode('utf-8')))

                # BOOLEAN
                elif isinstance(col.type, BOOLEAN):
                    filters.append('"{0}" = FALSE'.format(c.encode('utf-8')))

                else:
                    filters.append('TRIM(COALESCE(CAST("{0}" as Text), \'\')) = \'\''.format(c.encode('utf-8')))

            # ORDER BY: ASC
            elif operator0 == '+':
                order_by.append('"{0}" ASC'.format(c.encode('utf-8')))

            # ORDER BY: DESC
            elif operator0 == '-':
                order_by.append('"{0}" DESC'.format(c.encode('utf-8')))

            elif operator != '':
                # operator = <, >, <=, >=, =, <>, ==, !=

                # CONTAIN (=)
                if operator == '=':
                    # DATE
                    if isinstance(col.type, DATE):
                        operand_ = self._get_date(operand)
                        if operand_:
                            filters.append('"{0}" = \'{1}\''.format(c.encode('utf-8'), operand_))

                    # TIME
                    elif isinstance(col.type, TIME):
                        operand_ = self._get_time(operand)
                        if operand_:
                            filters.append('"{0}" = \'{1}\''.format(c.encode('utf-8'), operand_))

                    # DATETIME/TIMESTAMP
                    elif isinstance(col.type, DATETIME) or isinstance(col.type, TIMESTAMP):
                        operand_ = self._get_datetime(operand)
                        if operand_:
                            filters.append('"{0}" = \'{1}\''.format(c.encode('utf-8'), operand_))

                    # INTEGER, NUMERIC, BIGINT
                    elif isinstance(col.type, INTEGER) or isinstance(col.type, NUMERIC) or isinstance(col.type, BIGINT):
                        operand_ = self._get_number(operand)
                        if operand_:
                            filters.append('"{0}" = {1}'.format(c.encode('utf-8'), operand))

                    else:
                        args_ = (c.encode('utf-8'), self.no_accents(operand))
                        filters.append("UPPER(CAST(\"{0}\" as Text)) SIMILAR TO UPPER('%{1}%')".format(*args_))

                # NOT CONTAIN (<>)
                elif operator == '<>':
                    # DATE
                    if isinstance(col.type, DATE):
                        operand_ = self._get_date(operand)
                        if operand_:
                            filters.append('"{0}" != \'{1}\''.format(c.encode('utf-8'), operand_))

                    # TIME
                    elif isinstance(col.type, TIME):
                        operand_ = self._get_time(operand)
                        if operand_:
                            filters.append('"{0}" != \'{1}\''.format(c.encode('utf-8'), operand_))

                    # DATETIME/TIMESTAMP
                    elif isinstance(col.type, DATETIME) or isinstance(col.type, TIMESTAMP):
                        operand_ = self._get_datetime(operand)
                        if operand_:
                            filters.append('"{0}" != \'{1}\''.format(c.encode('utf-8'), operand_))

                    # INTEGER, NUMERIC, BIGINT
                    elif isinstance(col.type, INTEGER) or isinstance(col.type, NUMERIC) or isinstance(col.type, BIGINT):
                        operand_ = self._get_number(operand)
                        if operand_:
                            filters.append('"{0}" != {1}'.format(c.encode('utf-8'), operand))

                    else:
                        args_ = (c.encode('utf-8'), self.no_accents(operand))
                        filters.append("UPPER(CAST(\"{0}\" as Text)) NOT SIMILAR TO UPPER('%{1}%')".format(*args_))

                # EQUAL TO (==)
                elif operator == '==':
                    # DATE
                    if isinstance(col.type, DATE):
                        operand_ = self._get_date(operand)
                        if operand_:
                            filters.append('"{0}" = \'{1}\''.format(c.encode('utf-8'), operand_))

                    # TIME
                    elif isinstance(col.type, TIME):
                        operand_ = self._get_time(operand)
                        if operand_:
                            filters.append('"{0}" = \'{1}\''.format(c.encode('utf-8'), operand_))

                    # DATETIME/TIMESTAMP
                    elif isinstance(col.type, DATETIME) or isinstance(col.type, TIMESTAMP):
                        operand_ = self._get_datetime(operand)
                        if operand_:
                            filters.append('"{0}" = \'{1}\''.format(c.encode('utf-8'), operand_))

                    # INTEGER, NUMERIC, BIGINT
                    elif isinstance(col.type, INTEGER) or isinstance(col.type, NUMERIC) or isinstance(col.type, BIGINT):
                        operand_ = self._get_number(operand)
                        if operand_:
                            filters.append('"{0}" = {1}'.format(c.encode('utf-8'), operand))

                    else:
                        args_ = (c.encode('utf-8'), operand)
                        filters.append('UPPER(TRIM(COALESCE(CAST("{0}" as Text), \'\'))) = UPPER(\'{1}\')'.format(*args_))

                # NOT EQUAL TO (!=)
                elif operator == '!=':
                    # DATE
                    if isinstance(col.type, DATE):
                        operand_ = self._get_date(operand)
                        if operand_:
                            filters.append('"{0}" != \'{1}\''.format(c.encode('utf-8'), operand_))

                    # TIME
                    elif isinstance(col.type, TIME):
                        operand_ = self._get_time(operand)
                        if operand_:
                            filters.append('"{0}" != \'{1}\''.format(c.encode('utf-8'), operand_))

                    # DATETIME/TIMESTAMP
                    elif isinstance(col.type, DATETIME) or isinstance(col.type, TIMESTAMP):
                        operand_ = self._get_datetime(operand)
                        if operand_:
                            filters.append('"{0}" != \'{1}\''.format(c.encode('utf-8'), operand_))

                    # INTEGER, NUMERIC, BIGINT
                    elif isinstance(col.type, INTEGER) or isinstance(col.type, NUMERIC) or isinstance(col.type, BIGINT):
                        operand_ = self._get_number(operand)
                        if operand_:
                            filters.append('"{0}" != {1}'.format(c.encode('utf-8'), operand))

                    else:
                        args_ = (c.encode('utf-8'), operand)
                        filters.append('UPPER(TRIM(COALESCE(CAST("{0}" as Text), \'\'))) != UPPER(\'{1}\')'.format(*args_))

                # LESS THAN (<), LESS THAN OR EQUAL TO (<=), GREATER THAN (>), GREATER THAN OR EQUAL TO (>=)
                elif operator in ['<', '<=', '>', '>=']:
                    # DATE
                    if isinstance(col.type, DATE):
                        operand_ = self._get_date(operand)
                        if operand_:
                            filters.append('"{0}" {1} \'{2}\''.format(c.encode('utf-8'), operator, operand_))

                    # TIME
                    elif isinstance(col.type, TIME):
                        operand_ = self._get_time(operand)
                        if operand_:
                            filters.append('"{0}" {1} \'{2}\''.format(c.encode('utf-8'), operator, operand_))

                    # DATETIME/TIMESTAMP
                    elif isinstance(col.type, DATETIME) or isinstance(col.type, TIMESTAMP):
                        operand_ = self._get_datetime(operand)
                        if operand_:
                            filters.append('"{0}" {1} \'{2}\''.format(c.encode('utf-8'), operator, operand_))

                    elif isinstance(col.type, INTEGER) or isinstance(col.type, NUMERIC) or isinstance(col.type, BIGINT):
                        operand_ = self._get_number(operand)
                        if operand_:
                            filters.append('"{0}" {1} \'{2}\''.format(c.encode('utf-8'), operator, operand))

        filters_sql = None
        if len(filters) > 0:
            filters_sql = ('\nAND '.join(filters)).decode('utf-8')

        order_by_sql = None
        if len(order_by) > 0:
            order_by_sql = (', '.join(order_by)).decode('utf-8')

        return filters_sql, order_by_sql


class Search(object):

    def __init__(self, session, table_name, strtodatef=None):
        self.session = session
        self.table_name = table_name

        self.meta = MetaData(bind=self.session.bind)
        self.strtodatef = strtodatef
        self.sql = None
        self.order = ''

        if isinstance(self.table_name, Select):
            self.tbl = self.table_name

        else:
            self.tbl = Table(self.table_name, self.meta, autoload=True)

        self.from_ = self.tbl

    def apply_qry(self, q):

        if q:
            if isinstance(q, unicode):
                q = q.encode('utf-8')

            # process "q"
            qres = Busqueda(self.tbl, strtodatef=self.strtodatef)
            filters_sql, order_sql = qres.process(q)

            # apply search conditions
            self.sql = and_(filters_sql, self.sql)

            # apply order
            if order_sql:
                self.order = order_sql

    def and_(self, *cond):
        self.sql = and_(self.sql, *cond)

    def or_(self, cond):
        self.sql = or_(self.sql, cond)

    def join(self, *args, **kwargs):
        self.from_ = self.from_.join(*args, **kwargs)

    def outerjoin(self, *args, **kwargs):
        self.from_ = self.from_.outerjoin(*args, **kwargs)

    def apply_filters(self, filters):

        tbl = self.tbl

        # apply filters
        if filters:
            filters_tuple = (self.sql,)
            for f in filters:
                if len(f) > 2:
                    # (<field name>, <field value>, <operator>,)
                    # different
                    if f[2] == '!=':
                        filters_tuple += (tbl.c[f[0]] != f[1],)

                    # greater
                    elif f[2] == '>':
                        filters_tuple += (tbl.c[f[0]] > f[1],)

                    # greater or equal
                    elif f[2] == '>=':
                        filters_tuple += (tbl.c[f[0]] >= f[1],)

                    # less
                    elif f[2] == '<':
                        filters_tuple += (tbl.c[f[0]] < f[1],)

                    # less or equal
                    elif f[2] == '<=':
                        filters_tuple += (tbl.c[f[0]] <= f[1],)

                    # equal (and anything else...)
                    else:
                        filters_tuple += (tbl.c[f[0]] == f[1],)

                elif len(f) == 2:
                    # (<field name>, <field value>,)
                    # equal (by default)
                    filters_tuple += (tbl.c[f[0]] == f[1],)

                elif len(f) == 1:
                    filters_tuple += (f[0],)

            self.sql = and_(*filters_tuple)

    def __call__(self, rp=100, offset=0, collection=None, no_count=False, show_ids=False):
        """
        IN
          rp          <int>
          offset      <int>
          filters     [<tuple>, ...]
          collecion   <tuple> (<str>, <str>, <int>,)
          no_count    <bool> => False
          show_ids    <bool> => False

        OUT
          <DataSet>
        """

        sql = self.sql
        if collection:

            child_table_name = collection[0]
            child_attr = collection[1]
            parent_id = collection[2]

            child_attr_alias = '{0}$'.format(child_attr)
            if child_attr_alias in self.tbl.c:
                sql = and_(sql, self.tbl.c[child_attr_alias] == parent_id)

            else:
                child_table = alias(Table(child_table_name, self.meta, autoload=True))

                self.from_ = self.from_.\
                    join(child_table,
                         and_(child_table.c.id == self.tbl.c.id,
                              child_table.c[child_attr] == parent_id))

        # where
        if isinstance(self.tbl, Select):
            qry = self.tbl.where(sql)

        else:
            qry = select([self.tbl], from_obj=self.from_, whereclause=sql)

        # order by
        if self.order:
            qry = qry.order_by(self.order)

        return DataSet.procesar_resultado(self.session, qry, rp, offset, no_count=no_count, show_ids=show_ids)



# -*- coding: utf-8 -*-

from setuptools import setup

setup(name='Marcellus',
      version='1.1.1',
      author=u'León Domingo',
      author_email='leon@codevince.dev',
      description='Utilities for Quentin projects',
      #license=???,
      keywords='marcellus codevince web',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Topic :: Utilities',
      ],
      url='http://www.codevince.com',
      packages=['marcellus',
                'marcellus.postgres',
                'marcellus.templates',
                ],
      package_data={
          'marcellus.templates': ['*.xml'],
      },
      install_requires=[
          'psycopg2',
          'sqlalchemy',
          'lxml',
          'xlrd',
          'xlwt',
          'simplejson',
          'jinja2',
      ]
      )
